@@ -26,8 +26,8 @@
 
 #include "pglogical_compat.h"
 
-#define PGLOGICAL_VERSION "2.1.1"
-#define PGLOGICAL_VERSION_NUM 20101
+#define PGLOGICAL_VERSION "2.2.0"
+#define PGLOGICAL_VERSION_NUM 20200
 
 #define PGLOGICAL_MIN_PROTO_VERSION_NUM 1
 #define PGLOGICAL_MAX_PROTO_VERSION_NUM 1
@@ -77,4 +77,28 @@ extern void pglogical_create_sequence_state_record(Oid seqoid);
 extern void pglogical_drop_sequence_state_record(Oid seqoid);
 extern int64 sequence_get_last_value(Oid seqoid);
 
+#include "utils/memdebug.h"
+
+/*
+ * PostgreSQL exposes stubs for some Valgrind macros, but there are some
+ * others we use that aren't supported by Pg proper yet.
+ */
+#ifndef USE_VALGRIND
+#define VALGRIND_CHECK_VALUE_IS_DEFINED(v) do{} while(0)
+#define VALGRIND_DO_LEAK_CHECK do{} while(0)
+#define VALGRIND_DO_ADDED_LEAK_CHECK do{} while(0)
+#define VALGRIND_DO_CHANGED_LEAK_CHECK do{} while(0)
+#define VALGRIND_DO_QUICK_LEAK_CHECK do{} while(0)
+#define VALGRIND_DISABLE_ERROR_REPORTING do {} while (0)
+#define VALGRIND_ENABLE_ERROR_REPORTING do {} while (0)
+
+/*
+ * Gives us some error checking when no-op'd. pglogical uses this to report
+ * the worker type, etc, prefixed by PGLOGICAL:, in the Valgrind logs. We
+ * need to stub it out if we aren't using valgrind.
+ */
+pg_attribute_printf(1, 2) pg_attribute_unused() static inline void VALGRIND_PRINTF(const char *format, ...) {}
+
+#endif
+
 #endif /* PGLOGICAL_H */
