@@ -219,7 +219,7 @@ format_action_description(
 	}
 
 	appendStringInfo(si,
-			" in commit before %X/%X, xid %u commited at %s (action #%u)",
+			" in commit before %X/%X, xid %u committed at %s (action #%u)",
 			(uint32)(replorigin_session_origin_lsn>>32),
 			(uint32)replorigin_session_origin_lsn,
 			remote_xid,
@@ -300,6 +300,8 @@ handle_begin(StringInfo s)
 	replorigin_session_origin_lsn = commit_lsn;
 	remote_origin_id = InvalidRepOriginId;
 
+	VALGRIND_PRINTF("PGLOGICAL_APPLY: begin %u\n", remote_xid);
+
 	/* don't want the overhead otherwise */
 	if (apply_delay > 0)
 	{
@@ -467,6 +469,8 @@ handle_commit(StringInfo s)
 		proc_exit(0);
 	}
 
+	VALGRIND_PRINTF("PGLOGICAL_APPLY: commit %u\n", remote_xid);
+
 	xact_action_counter = 0;
 	remote_xid = InvalidTransactionId;
 
@@ -905,8 +909,9 @@ handle_table_sync(QueuedMessage *queued_message)
 	create_local_sync_status(newsync);
 
 	oldcontext = MemoryContextSwitchTo(TopMemoryContext);
-	SyncingTables = lappend(SyncingTables, newsync);
 	MemoryContextSwitchTo(oldcontext);
+
+	MyApplyWorker->sync_pending = true;
 }
 
 /*
@@ -1092,6 +1097,8 @@ replication_handler(StringInfo s)
 	errcallback.previous = error_context_stack;
 	error_context_stack = &errcallback;
 
+	Assert(CurrentMemoryContext == MessageContext);
+
 	switch (action)
 	{
 		/* BEGIN */
@@ -1130,8 +1137,21 @@ replication_handler(StringInfo s)
 			elog(ERROR, "unknown action of type %c", action);
 	}
 
+	Assert(CurrentMemoryContext == MessageContext);
+
 	if (error_context_stack == &errcallback)
 		error_context_stack = errcallback.previous;
+
+	if (action == 'C')
+	{
+		/*
+		 * We clobber MessageContext on commit. It doesn't matter much when we
+		 * do it so long as we do so periodically, to prevent the context from
+		 * growing too much. We might want to clean it up even 'n'th message
+		 * too, but that adds testing burden and isn't done for now.
+		 */
+		MemoryContextReset(MessageContext);
+	}
 }
 
 /*
@@ -1292,8 +1312,11 @@ apply_work(PGconn *streamConn)
 										   ALLOCSET_DEFAULT_INITSIZE,
 										   ALLOCSET_DEFAULT_MAXSIZE);
 
+	MemoryContextSwitchTo(MessageContext);
+
 	/* mark as idle, before starting to loop */
 	pgstat_report_activity(STATE_IDLE, NULL);
+	Assert(CurrentMemoryContext == MessageContext);
 
 	while (!got_SIGTERM)
 	{
@@ -1313,31 +1336,31 @@ apply_work(PGconn *streamConn)
 
 		ResetLatch(&MyProc->procLatch);
 
-		MemoryContextSwitchTo(MessageContext);
+		Assert(CurrentMemoryContext == MessageContext);
 
 		/* emergency bailout if postmaster has died */
 		if (rc & WL_POSTMASTER_DEATH)
 			proc_exit(1);
 
+		if (rc & WL_SOCKET_READABLE)
+			PQconsumeInput(applyconn);
+
 		if (PQstatus(applyconn) == CONNECTION_BAD)
 		{
 			elog(ERROR, "connection to other side has died");
 		}
 
-		if (rc & WL_SOCKET_READABLE)
-			PQconsumeInput(applyconn);
+		Assert(CurrentMemoryContext == MessageContext);
 
 		for (;;)
 		{
 			if (got_SIGTERM)
 				break;
 
-			if (copybuf != NULL)
-			{
-				PQfreemem(copybuf);
-				copybuf = NULL;
-			}
+			/* We must not have fallen out of MessageContext by accident */
+			Assert(CurrentMemoryContext == MessageContext);
 
+			Assert(copybuf == NULL);
 			r = PQgetCopyData(applyconn, &copybuf, 1);
 
 			if (r == -1)
@@ -1361,12 +1384,15 @@ apply_work(PGconn *streamConn)
 				int c;
 				StringInfoData s;
 
-				MemoryContextSwitchTo(MessageContext);
-
-				initStringInfo(&s);
+				/*
+				 * We're using a StringInfo to wrap existing data here, as a
+				 * cursor. We init it manually to avoid a redundant allocation.
+				 */
+				memset(&s, 0, sizeof(StringInfoData));
 				s.data = copybuf;
 				s.len = r;
 				s.maxlen = -1;
+				s.cursor = 0;
 
 				c = pq_getmsgbyte(&s);
 
@@ -1401,7 +1427,17 @@ apply_work(PGconn *streamConn)
 								  reply_requested);
 				}
 				/* other message types are purposefully ignored */
+
+				/* copybuf is malloc'd not palloc'd */
+				if (copybuf != NULL)
+				{
+					PQfreemem(copybuf);
+					copybuf = NULL;
+				}
 			}
+
+			/* We must not have fallen out of MessageContext by accident */
+			Assert(CurrentMemoryContext == MessageContext);
 		}
 
 		/* confirm all writes at once */
@@ -1409,9 +1445,21 @@ apply_work(PGconn *streamConn)
 
 		if (!in_remote_transaction)
 			process_syncing_tables(last_received);
+		
+		/* We must not have switched out of MessageContext by mistake */
+		Assert(CurrentMemoryContext == MessageContext);
 
 		/* Cleanup the memory. */
 		MemoryContextResetAndDeleteChildren(MessageContext);
+
+		/*
+		 * Only do a leak check if we're between txns; we don't want lots of
+		 * noise due to resources that only exist in a txn.
+		 */
+		if (!IsTransactionState())
+		{
+			VALGRIND_DO_ADDED_LEAK_CHECK;
+		}
 	}
 }
 
@@ -1591,6 +1639,9 @@ process_syncing_tables(XLogRecPtr end_lsn)
 {
 	ListCell	   *lc;
 
+	Assert(CurrentMemoryContext == MessageContext);
+	Assert(!IsTransactionState());
+
 	/* First check if we need to update the cached information. */
 	if (MyApplyWorker->sync_pending)
 	{
@@ -1598,6 +1649,7 @@ process_syncing_tables(XLogRecPtr end_lsn)
 		MyApplyWorker->sync_pending = false;
 		reread_unsynced_tables(MyApplyWorker->subid);
 		CommitTransactionCommand();
+		MemoryContextSwitchTo(MessageContext);
 	}
 
 	/* Process currently pending sync tables. */
@@ -1637,6 +1689,7 @@ process_syncing_tables(XLogRecPtr end_lsn)
 			else
 				memcpy(sync, newsync, sizeof(PGLogicalSyncStatus));
 			CommitTransactionCommand();
+			MemoryContextSwitchTo(MessageContext);
 
 			if (sync->status == SYNC_STATUS_SYNCWAIT)
 			{
@@ -1661,6 +1714,7 @@ process_syncing_tables(XLogRecPtr end_lsn)
 										  sync->status,
 										  sync->statuslsn);
 					CommitTransactionCommand();
+					MemoryContextSwitchTo(MessageContext);
 
 					if (pglogical_worker_running(worker))
 						SetLatch(&worker->proc->procLatch);
@@ -1690,6 +1744,7 @@ process_syncing_tables(XLogRecPtr end_lsn)
 									  sync->status,
 									  sync->statuslsn);
 				CommitTransactionCommand();
+				MemoryContextSwitchTo(MessageContext);
 			}
 
 			/* Ready? Remove it from local cache. */
@@ -1734,6 +1789,8 @@ process_syncing_tables(XLogRecPtr end_lsn)
 			break;
 		}
 	}
+
+	Assert(CurrentMemoryContext == MessageContext);
 }
 
 static void
@@ -1792,15 +1849,11 @@ pglogical_apply_main(Datum main_arg)
 
 	/* Establish signal handlers. */
 	pqsignal(SIGTERM, handle_sigterm);
-	BackgroundWorkerUnblockSignals();
 
 	/* Attach to dsm segment. */
 	Assert(CurrentResourceOwner == NULL);
 	CurrentResourceOwner = ResourceOwnerCreate(NULL, "pglogical apply");
 
-	/* Connect to our database. */
-	BackgroundWorkerInitializeConnectionByOid(MyPGLogicalWorker->dboid, InvalidOid);
-
 	/* Load correct apply API. */
 	if (pglogical_use_spi)
 	{
@@ -1899,6 +1952,14 @@ pglogical_apply_main(Datum main_arg)
 
 	CommitTransactionCommand();
 
+	/*
+	 * Do an initial leak check with reporting off; we don't want to see
+	 * these results, just the later output from ADDED leak checks.
+	 */
+	VALGRIND_DISABLE_ERROR_REPORTING;
+	VALGRIND_DO_LEAK_CHECK;
+	VALGRIND_ENABLE_ERROR_REPORTING;
+
 	apply_work(streamConn);
 
 	PQfinish(streamConn);
