@@ -33,6 +33,7 @@
 #include "utils/memutils.h"
 #include "utils/rel.h"
 
+#include "pglogical.h"
 #include "pglogical_output_config.h"
 #include "pglogical_executor.h"
 #include "pglogical_node.h"
@@ -74,12 +75,13 @@ typedef struct PGLRelMetaCacheEntry
 
 #define RELMETACACHE_INITIAL_SIZE 128
 static HTAB *RelMetaCache = NULL;
+static MemoryContext RelMetaCacheContext = NULL;
 static int InvalidRelMetaCacheCnt = 0;
 
 static void relmetacache_init(MemoryContext decoding_context);
 static PGLRelMetaCacheEntry *relmetacache_get_relation(PGLogicalOutputData *data,
 													   Relation rel);
-static void relmetacache_destroy(void);
+static void relmetacache_flush(void);
 static void relmetacache_prune(void);
 
 static void pglReorderBufferCleanSerializedTXNs(const char *slotname);
@@ -165,8 +167,9 @@ pg_decode_startup(LogicalDecodingContext * ctx, OutputPluginOptions *opt,
 {
 	PGLogicalOutputData	   *data = palloc0(sizeof(PGLogicalOutputData));
 
+	/* Short lived memory context for individual messages */
 	data->context = AllocSetContextCreate(ctx->context,
-										  "pglogical conversion context",
+										  "pglogical output msg context",
 										  ALLOCSET_DEFAULT_MINSIZE,
 										  ALLOCSET_DEFAULT_INITSIZE,
 										  ALLOCSET_DEFAULT_MAXSIZE);
@@ -368,6 +371,13 @@ pg_decode_startup(LogicalDecodingContext * ctx, OutputPluginOptions *opt,
 
 		relmetacache_init(ctx->context);
 	}
+
+	/* So we can identify the process type in Valgrind logs */
+	VALGRIND_PRINTF("PGLOGICAL: pglogical worker output_plugin\n");
+	/* For incremental leak checking */
+	VALGRIND_DISABLE_ERROR_REPORTING;
+	VALGRIND_DO_LEAK_CHECK;
+	VALGRIND_ENABLE_ERROR_REPORTING;
 }
 
 /*
@@ -378,6 +388,11 @@ pg_decode_begin_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn)
 {
 	PGLogicalOutputData* data = (PGLogicalOutputData*)ctx->output_plugin_private;
 	bool send_replication_origin = data->forward_changeset_origins;
+	MemoryContext old_ctx;
+
+	old_ctx = MemoryContextSwitchTo(data->context);
+
+	VALGRIND_DO_ADDED_LEAK_CHECK;
 
 	if (!startup_message_sent)
 		send_startup_message(ctx, data, false /* can't be last message */);
@@ -415,6 +430,9 @@ pg_decode_begin_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn)
 #endif
 
 	OutputPluginWrite(ctx, true);
+
+	Assert(CurrentMemoryContext == data->context);
+	MemoryContextSwitchTo(old_ctx);
 }
 
 /*
@@ -425,6 +443,9 @@ pg_decode_commit_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 					 XLogRecPtr commit_lsn)
 {
 	PGLogicalOutputData* data = (PGLogicalOutputData*)ctx->output_plugin_private;
+	MemoryContext old_ctx;
+
+	old_ctx = MemoryContextSwitchTo(data->context);
 
 	OutputPluginPrepareWrite(ctx, true);
 	data->api->write_commit(ctx->out, data, txn, commit_lsn);
@@ -436,6 +457,12 @@ pg_decode_commit_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	 * at the moment.
 	 */
 	relmetacache_prune();
+
+	Assert(CurrentMemoryContext == data->context);
+	MemoryContextSwitchTo(old_ctx);
+	MemoryContextReset(data->context);
+
+	VALGRIND_DO_ADDED_LEAK_CHECK;
 }
 
 static bool
@@ -629,13 +656,13 @@ pg_decode_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	MemoryContext	old;
 	Bitmapset	   *att_list = NULL;
 
+	/* Avoid leaking memory by using and resetting our own context */
+	old = MemoryContextSwitchTo(data->context);
+
 	/* First check the table filter */
 	if (!pglogical_change_filter(data, relation, change, &att_list))
 		return;
 
-	/* Avoid leaking memory by using and resetting our own context */
-	old = MemoryContextSwitchTo(data->context);
-
 	/*
 	 * If the protocol wants to write relation information and the client
 	 * isn't known to have metadata cached for this relation already,
@@ -696,6 +723,7 @@ pg_decode_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	}
 
 	/* Cleanup */
+	Assert(CurrentMemoryContext == data->context);
 	MemoryContextSwitchTo(old);
 	MemoryContextReset(data->context);
 }
@@ -760,7 +788,9 @@ send_startup_message(LogicalDecodingContext *ctx,
 static void
 pg_decode_shutdown(LogicalDecodingContext * ctx)
 {
-	relmetacache_destroy();
+	relmetacache_flush();
+
+	VALGRIND_PRINTF("PGLOGICAL: output plugin shutdown\n");
 
 	/*
 	 * no need to delete data->context as it's child of ctx->context which
@@ -820,13 +850,19 @@ relmetacache_init(MemoryContext decoding_context)
 	{
 		MemoryContext old_ctxt;
 
+		RelMetaCacheContext = AllocSetContextCreate(TopMemoryContext,
+											  "pglogical output relmetacache",
+											  ALLOCSET_DEFAULT_MINSIZE,
+											  ALLOCSET_DEFAULT_INITSIZE,
+											  ALLOCSET_DEFAULT_MAXSIZE);
+
 		/* Make a new hash table for the cache */
 		hash_flags = HASH_ELEM | HASH_CONTEXT;
 
 		MemSet(&ctl, 0, sizeof(ctl));
 		ctl.keysize = sizeof(Oid);
 		ctl.entrysize = sizeof(struct PGLRelMetaCacheEntry);
-		ctl.hcxt = TopMemoryContext;
+		ctl.hcxt = RelMetaCacheContext;
 
 #if PG_VERSION_NUM >= 90500
 		hash_flags |= HASH_BLOBS;
@@ -835,7 +871,7 @@ relmetacache_init(MemoryContext decoding_context)
 		hash_flags |= HASH_FUNCTION;
 #endif
 
-		old_ctxt = MemoryContextSwitchTo(TopMemoryContext);
+		old_ctxt = MemoryContextSwitchTo(RelMetaCacheContext);
 		RelMetaCache = hash_create("pglogical relation metadata cache",
 								   RELMETACACHE_INITIAL_SIZE,
 								   &ctl, hash_flags);
@@ -866,7 +902,7 @@ relmetacache_get_relation(struct PGLogicalOutputData *data,
 	MemoryContext old_mctx;
 
 	/* Find cached function info, creating if not found */
-	old_mctx = MemoryContextSwitchTo(TopMemoryContext);
+	old_mctx = MemoryContextSwitchTo(RelMetaCacheContext);
 	hentry = (struct PGLRelMetaCacheEntry*) hash_search(RelMetaCache,
 										 (void *)(&RelationGetRelid(rel)),
 										 HASH_ENTER, &found);
@@ -889,9 +925,13 @@ relmetacache_get_relation(struct PGLogicalOutputData *data,
 
 /*
  * Flush the relation metadata cache at the end of a decoding session.
+ *
+ * We cannot truly destroy the cache because it may be referenced by later
+ * relcache invalidation callbacks after the end of a SQL-level decoding
+ * session.
  */
 static void
-relmetacache_destroy(void)
+relmetacache_flush(void)
 {
 	HASH_SEQ_STATUS status;
 	struct PGLRelMetaCacheEntry *hentry;
