@@ -157,46 +157,56 @@ restore_structure(PGLogicalSubscription *sub, const char *srcfile,
 						command.data)));
 }
 
-
 /*
- * Ensure slot exists.
+ * Create slot and get the exported snapshot.
+ *
+ * This will try to recreate slot if already exists and not active.
+ *
+ * The reported LSN is the confirmed flush LSN at the point the slot reached
+ * consistency and exported its snapshot.
  */
 static char *
-ensure_replication_slot_snapshot(PGconn *origin_conn, char *slot_name,
-								 bool use_failover_slot, XLogRecPtr *lsn)
+ensure_replication_slot_snapshot(PGconn *sql_conn, PGconn *repl_conn,
+								 char *slot_name, bool use_failover_slot,
+								 XLogRecPtr *lsn)
 {
 	PGresult	   *res;
 	StringInfoData	query;
 	char		   *snapshot;
 
+retry:
 	initStringInfo(&query);
 
 	appendStringInfo(&query, "CREATE_REPLICATION_SLOT \"%s\" LOGICAL %s%s",
 					 slot_name, "pglogical_output",
 					 use_failover_slot ? " FAILOVER" : "");
 
-	res = PQexec(origin_conn, query.data);
 
-	/* TODO: check and handle already existing slot. */
+	res = PQexec(repl_conn, query.data);
+
 	if (PQresultStatus(res) != PGRES_TUPLES_OK)
 	{
-		char	*err = PQresultErrorMessage(res);
+		const char *sqlstate = PQresultErrorField(res, PG_DIAG_SQLSTATE);
 
 		/*
-		 * Report more pleasant message on snapshot too large
-		 * (there is not ERRCODE for it).
-		 *
-		 * Note that retry is not hanled by us but by the fact that this
-		 * function is only called when sync state is INIT.
+		 * If our slot already exist but is not used, it's leftover from
+		 * previous unsucessful attempt to synchronize table, try dropping
+		 * it and recreating.
 		 */
-		if (strstr(err, "snapshot too large"))
-			ereport(ERROR,
-					(errmsg("could not start synchronization, will retry later"),
-					 errdetail("recieved \"snapshot too large\" from provider")));
-		else
-			elog(FATAL, "could not send replication command \"%s\": status %s: %s\n",
-				 query.data,
-				 PQresStatus(PQresultStatus(res)), err);
+		if (sqlstate &&
+			strcmp(sqlstate, "42710" /*ERRCODE_DUPLICATE_OBJECT*/) == 0 &&
+			!pglogical_remote_slot_active(sql_conn, slot_name))
+		{
+			pfree(query.data);
+			PQclear(res);
+
+			pglogical_drop_remote_slot(sql_conn, slot_name);
+
+			goto retry;
+		}
+
+		elog(ERROR, "could not create replication slot on provider: %s\n",
+			 PQresultErrorMessage(res));
 	}
 
 	*lsn = DatumGetLSN(DirectFunctionCall1Coll(pg_lsn_in, InvalidOid,
@@ -750,15 +760,16 @@ pglogical_sync_subscription(PGLogicalSubscription *sub)
 			pglogical_remote_function_exists(origin_conn, "pg_catalog",
 											 "pg_create_logical_replication_slot",
 											 3);
-		PQfinish(origin_conn);
-
 		origin_conn_repl = pglogical_connect_replica(sub->origin_if->dsn,
 													 sub->name, "snap");
 
-		snapshot = ensure_replication_slot_snapshot(origin_conn_repl,
+		snapshot = ensure_replication_slot_snapshot(origin_conn,
+													origin_conn_repl,
 													sub->slot_name,
 													use_failover_slot, &lsn);
 
+		PQfinish(origin_conn);
+
 		PG_ENSURE_ERROR_CLEANUP(pglogical_sync_worker_cleanup_error_cb,
 								PointerGetDatum(sub));
 		{
@@ -908,7 +919,7 @@ char
 pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table,
 					 XLogRecPtr *status_lsn)
 {
-	PGconn	   *origin_conn_repl;
+	PGconn	   *origin_conn_repl, *origin_conn;
 	RepOriginId	originid;
 	char	   *snapshot;
 	PGLogicalSyncStatus	   *sync;
@@ -942,9 +953,11 @@ pglogical_sync_table(PGLogicalSubscription *sub, RangeVar *table,
 	origin_conn_repl = pglogical_connect_replica(sub->origin_if->dsn,
 												 sub->name, "copy");
 
-	snapshot = ensure_replication_slot_snapshot(origin_conn_repl,
+	origin_conn = pglogical_connect(sub->origin_if->dsn, sub->name, "copy_slot");
+	snapshot = ensure_replication_slot_snapshot(origin_conn, origin_conn_repl,
 												sub->slot_name, false,
 												status_lsn);
+	PQfinish(origin_conn);
 
 	/* Make sure we cleanup the slot if something goes wrong. */
 	PG_ENSURE_ERROR_CLEANUP(pglogical_sync_worker_cleanup_error_cb,
@@ -1045,15 +1058,11 @@ pglogical_sync_main(Datum main_arg)
 
 	/* Establish signal handlers. */
 	pqsignal(SIGTERM, handle_sigterm);
-	BackgroundWorkerUnblockSignals();
 
 	/* Attach to dsm segment. */
 	Assert(CurrentResourceOwner == NULL);
 	CurrentResourceOwner = ResourceOwnerCreate(NULL, "pglogical sync");
 
-	/* Connect to our database. */
-	BackgroundWorkerInitializeConnectionByOid(MyPGLogicalWorker->dboid, InvalidOid);
-
 	/* Setup synchronous commit according to the user's wishes */
 	SetConfigOption("synchronous_commit",
 					pglogical_synchronous_commit ? "local" : "off",
@@ -1416,6 +1425,45 @@ drop_table_sync_status(const char *nspname, const char *relname)
 
 }
 
+/* Remove table sync status record from catalog. */
+void
+drop_table_sync_status_for_sub(Oid subid, const char *nspname,
+							   const char *relname)
+{
+	RangeVar	   *rv;
+	Relation		rel;
+	SysScanDesc		scan;
+	HeapTuple		tuple;
+	ScanKeyData		key[3];
+
+	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
+	rel = heap_openrv(rv, RowExclusiveLock);
+
+	ScanKeyInit(&key[0],
+				Anum_sync_subid,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(subid));
+	ScanKeyInit(&key[1],
+				Anum_sync_nspname,
+				BTEqualStrategyNumber, F_NAMEEQ,
+				CStringGetDatum(nspname));
+	ScanKeyInit(&key[2],
+				Anum_sync_relname,
+				BTEqualStrategyNumber, F_NAMEEQ,
+				CStringGetDatum(relname));
+
+	scan = systable_beginscan(rel, 0, true, NULL, 3, key);
+
+	/* Remove the tuples. */
+	while (HeapTupleIsValid(tuple = systable_getnext(scan)))
+		simple_heap_delete(rel, &tuple->t_self);
+
+	/* Cleanup. */
+	systable_endscan(scan);
+	heap_close(rel, RowExclusiveLock);
+
+}
+
 /* Get the sync status for a table. */
 PGLogicalSyncStatus *
 get_table_sync_status(Oid subid, const char *nspname, const char *relname,
@@ -1511,6 +1559,46 @@ get_unsynced_tables(Oid subid)
 	return res;
 }
 
+/* Get the sync status for all tables known to subscription. */
+List *
+get_subscription_tables(Oid subid)
+{
+	PGLogicalSyncStatus	   *sync;
+	RangeVar	   *rv;
+	Relation		rel;
+	SysScanDesc		scan;
+	HeapTuple		tuple;
+	ScanKeyData		key[1];
+	List		   *res = NIL;
+	TupleDesc		tupDesc;
+
+	rv = makeRangeVar(EXTENSION_NAME, CATALOG_LOCAL_SYNC_STATUS, -1);
+	rel = heap_openrv(rv, RowExclusiveLock);
+	tupDesc = RelationGetDescr(rel);
+
+	ScanKeyInit(&key[0],
+				Anum_sync_subid,
+				BTEqualStrategyNumber, F_OIDEQ,
+				ObjectIdGetDatum(subid));
+
+	scan = systable_beginscan(rel, 0, true, NULL, 1, key);
+
+	while (HeapTupleIsValid(tuple = systable_getnext(scan)))
+	{
+		if (heap_attisnull(tuple, Anum_sync_nspname) &&
+			heap_attisnull(tuple, Anum_sync_relname))
+			continue;
+
+		sync = syncstatus_fromtuple(tuple, tupDesc);
+		res = lappend(res, sync);
+	}
+
+	systable_endscan(scan);
+	heap_close(rel, RowExclusiveLock);
+
+	return res;
+}
+
 /* Set the sync status for a table. */
 void
 set_table_sync_status(Oid subid, const char *nspname, const char *relname,
@@ -1575,15 +1663,22 @@ set_table_sync_status(Oid subid, const char *nspname, const char *relname,
  *
  * We also exit if the worker is no longer recognized as sync worker as
  * that means something bad happened to it.
+ *
+ * Care with allocations is required here since it typically runs
+ * in TopMemoryContext.
  */
 bool
 wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
 							char desired_state, XLogRecPtr *lsn)
 {
 	int rc;
+	MemoryContext old_ctx = CurrentMemoryContext;
+	bool ret = false;
 
 	*lsn = InvalidXLogRecPtr;
 
+	Assert(!IsTransactionState());
+
 	while (!got_SIGTERM)
 	{
 		PGLogicalWorker		   *worker;
@@ -1594,22 +1689,24 @@ wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
 		if (!sync)
 		{
 			CommitTransactionCommand();
-			return false;
+			break;
 		}
 		if (sync->status == desired_state)
 		{
 			*lsn = sync->statuslsn;
 			CommitTransactionCommand();
-			return true;
+			ret = true;
+			break;
 		}
 		CommitTransactionCommand();
+		(void) MemoryContextSwitchTo(old_ctx);
 
 		/* Check if the worker is still alive - no point waiting if it died. */
 		LWLockAcquire(PGLogicalCtx->lock, LW_EXCLUSIVE);
 		worker = pglogical_sync_find(MyDatabaseId, subid, nspname, relname);
 		LWLockRelease(PGLogicalCtx->lock);
 		if (!worker)
-			return false;
+			break;
 
 		rc = WaitLatch(&MyProc->procLatch,
 					   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH,
@@ -1622,7 +1719,8 @@ wait_for_sync_status_change(Oid subid, char *nspname, char *relname,
 			proc_exit(1);
 	}
 
-	return false; /* Silence compiler. */
+	(void) MemoryContextSwitchTo(old_ctx);
+	return ret;
 }
 
 /*
