@@ -14,8 +14,12 @@
 
 #include "miscadmin.h"
 
+#include "libpq/libpq-be.h"
+
 #include "access/xact.h"
 
+#include "commands/dbcommands.h"
+
 #include "storage/ipc.h"
 #include "storage/proc.h"
 #include "storage/procsignal.h"
@@ -28,6 +32,11 @@
 #include "pglogical_sync.h"
 #include "pglogical_worker.h"
 
+typedef struct signal_workers_arg
+{
+	Oid		subid;
+	bool	kill;
+} signal_workers_arg;
 
 volatile sig_atomic_t	got_SIGTERM = false;
 
@@ -142,8 +151,8 @@ pglogical_worker_register(PGLogicalWorker *worker)
 		snprintf(bgw.bgw_function_name, BGW_MAXLEN,
 				 "pglogical_sync_main");
 		snprintf(bgw.bgw_name, BGW_MAXLEN,
-				 "pglogical sync %s %u:%u",
-				 NameStr(worker->worker.sync.relname),
+				 "pglogical sync %*s %u:%u", NAMEDATALEN - 37,
+				 shorten_hash(NameStr(worker->worker.sync.relname), NAMEDATALEN - 37),
 				 worker->dboid, worker->worker.sync.apply.subid);
 	}
 	else
@@ -292,6 +301,8 @@ pglogical_worker_attach(int slot, PGLogicalWorkerType type)
 	Assert(slot >= 0);
 	Assert(slot < PGLogicalCtx->total_workers);
 
+	MyProcPort = (Port *) calloc(1, sizeof(Port));
+
 #if PG_VERSION_NUM < 90600
 	set_latch_on_sigusr1 = true;
 #endif
@@ -310,11 +321,40 @@ pglogical_worker_attach(int slot, PGLogicalWorkerType type)
 		 pglogical_worker_type_name(type), MyProcPid, slot,
 		 MyPGLogicalWorkerGeneration);
 
+	/*
+	 * So we can find workers in valgrind output, send a Valgrind client
+	 * request to print to the Valgrind log.
+	 */
+	VALGRIND_PRINTF("PGLOGICAL: pglogical worker %s (%s)\n",
+		pglogical_worker_type_name(type),
+		MyBgworkerEntry->bgw_name);
+
 	LWLockRelease(PGLogicalCtx->lock);
 
 	/* Make it easy to identify our processes. */
 	SetConfigOption("application_name", MyBgworkerEntry->bgw_name,
 					PGC_USERSET, PGC_S_SESSION);
+
+	/* Establish signal handlers. */
+	BackgroundWorkerUnblockSignals();
+
+	/* Make it easy to identify our processes. */
+	SetConfigOption("application_name", MyBgworkerEntry->bgw_name,
+					PGC_USERSET, PGC_S_SESSION);
+
+	/* Connect to database if needed. */
+	if (MyPGLogicalWorker->dboid != InvalidOid)
+	{
+		MemoryContext oldcontext;
+
+		BackgroundWorkerInitializeConnectionByOid(MyPGLogicalWorker->dboid,
+												  InvalidOid);
+		StartTransactionCommand();
+		oldcontext = MemoryContextSwitchTo(TopMemoryContext);
+		MyProcPort->database_name = pstrdup(get_database_name(MyPGLogicalWorker->dboid));
+		MemoryContextSwitchTo(oldcontext);
+		CommitTransactionCommand();
+	}
 }
 
 /*
@@ -341,6 +381,9 @@ pglogical_worker_detach(bool crash)
 		 MyPGLogicalWorkerGeneration,
 		 crash ? "exiting with error" : "detaching cleanly");
 
+	VALGRIND_PRINTF("PGLOGICAL: worker detaching, unclean=%d\n",
+		crash);
+
 	/*
 	 * If we crashed we need to report it.
 	 *
@@ -528,11 +571,17 @@ signal_worker_xact_callback(XactEvent event, void *arg)
 
 				if (arg != NULL)
 				{
-					Oid					subid = *(Oid *) arg;
+					signal_workers_arg *swa = (signal_workers_arg *) arg;
 					PGLogicalWorker	   *apply;
 
-					apply = pglogical_apply_find(MyDatabaseId, subid);
-					pglogical_worker_kill(apply);
+					apply = pglogical_apply_find(MyDatabaseId, swa->subid);
+					if (swa->kill)
+						pglogical_worker_kill(apply);
+					else if (pglogical_worker_running(apply))
+					{
+						apply->worker.apply.sync_pending = true;
+						SetLatch(&apply->proc->procLatch);
+					}
 				}
 
 				PGLogicalCtx->subscriptions_changed = true;
@@ -563,7 +612,7 @@ signal_worker_xact_callback(XactEvent event, void *arg)
  * Enqueue singal for supervisor/manager at COMMIT.
  */
 void
-pglogical_subscription_changed(Oid subid)
+pglogical_subscription_changed(Oid subid, bool kill)
 {
 	if (!xacthook_signal_workers)
 	{
@@ -571,8 +620,12 @@ pglogical_subscription_changed(Oid subid)
 
 		if (OidIsValid(subid))
 		{
-			arg = MemoryContextAlloc(TopTransactionContext, sizeof(subid));
-			memcpy(arg, &subid, sizeof(subid));
+			signal_workers_arg swa;
+			arg = MemoryContextAlloc(TopTransactionContext,
+									 sizeof(signal_workers_arg));
+			swa.subid = subid;
+			swa.kill = kill;
+			memcpy(arg, &swa, sizeof(signal_workers_arg));
 		}
 
 		RegisterXactCallback(signal_worker_xact_callback, arg);
